module Main where

import Control.Monad
import Data.IORef
import Data.Maybe

import System.Random

import qualified Graphics.UI.Threepenny as UI
import Graphics.UI.Threepenny.Core

import Minesweeper
import Autosolver

main :: IO ()
main = startGUI defaultConfig {jsStatic = Just "view"} setup

setup :: Window -> UI ()
setup w = void $ do
    return w # set title "Minesweeper"

    UI.addStyleSheet w "bootstrap.min.css"
    UI.addStyleSheet w "minesweeper.css"

    rng <- liftIO newStdGen
    let b = createBoard 20 0.08 rng
    iob <- liftIO $ newIORef b

    getBody w #+ [
        UI.div #. "container" #+ [
            UI.div #. "row" #+ [
                UI.h1 #+ [string "Minesweeper"],
                UI.br,
                UI.p #+ [string "Jack Harley jharley@tcd.ie"]
            ],
            UI.div #. "row" #+ [
                UI.div # set UI.id_ "gameCont" #+ [mkElement "table" # set UI.id_ "table" #+ rows iob b 0],
                UI.div # set UI.id_ "infoCont" #+ [
                    UI.p #+ [string "Instructions: Click on a square to uncover it. Right click a square to flag it."],
                    UI.p #+ [string "Flagged squares will turn yellow. If you hit a mine all mines will instantly be revealed as red squares."],
                    UI.p #+ [string "You win the game once you have uncovered all squares that do not have mines. If this occurs, the entire board will turn green (except the bomb) to indicate your win!"],
                    UI.p #+ [string "At any time, you can refresh the page to start a new game."],
                    UI.p #+ [string "Good luck!"]
                ] 
            ],
            UI.div #. "row" #+ [
                UI.div #. "card" #+ [
                    UI.div #. "card-header" #+ [string "Autoplayer"],
                    UI.div #. "card-body" #+ [
                        UI.p #+ [string "Not sure what to do? Click the Autoplay button below to let the computer make a move for you!"],
                        UI.p #+ [autoPlayButton iob]
                    ]
                ]
            ]
        ],

        -- include custom JS at end of body to stop right clicks opening a menu
        mkElement "script" # set (attr "src") "/static/custom.js"]

    where
        rows iob b r | r < size b = (mkElement "tr" #+ cells iob b r 0) : rows iob b (r+1)
                     | otherwise = []

        cells iob b r c | c < size b = cell iob b (r,c) : cells iob b r (c+1)
                        | otherwise = []
        
        cell iob b (r,c) = do
            cell <- mkElement "td" #. (squareBgColour b (r,c) ++ " " ++ squareTextColour b (r,c)) 
                                   #+ [string $ squareAscii b (r,c)]

            on UI.click cell $ \_ -> do
                --liftIO $ putStrLn $ "Click (" ++ show r ++ "," ++ show c ++ ")"
                liftIO $ modifyIORef' iob $ \oldB -> uncover oldB (r,c)
                refresh iob

            on UI.contextmenu cell $ \_ -> do
                --liftIO $ putStrLn $ "Right Click (" ++ show r ++ "," ++ show c ++ ")"
                liftIO $ modifyIORef' iob $ \oldB -> flag oldB (r,c)
                refresh iob

            return cell

        -- auto play move button
        autoPlayButton iob = do
            button <- UI.button #. "btn btn-primary" #+ [string "Autoplay"]

            on UI.click button $ \_ -> do
                liftIO $ modifyIORef' iob $ \b -> autoplay b
                refresh iob

            return button

        -- refresh the board on screen (rerender)
        refresh iob = do
            b <- liftIO $ readIORef iob
                
            table <- getElementById w "table"
                
            cont <- getElementById w "gameCont"
            let cont' = return $ fromJust cont
            cont' #+ [mkElement "table" # set UI.id_ "table" #+ rows iob b 0]

            when (isJust table) $ delete (fromJust table)

            -- For some reason, ocassionally threepenny will fail to render the table after a change.
            -- Despite extensive debugging I cannot determine why, I believe there may be some type of
            -- bug in threepenny causing this, the underlying data structures all appear fine and
            -- simply forcing a second refresh always fixes it.
            -- Therefore: to workaround, we'll check if the render failed and if so, render again

            newTable <- getElementById w "table"
            when (isNothing newTable) $ do
                liftIO $ putStrLn "Render failed, triggering repeat"
                refresh iob